from typing import List, Tuple

from Bio.motifs import create
from Bio.SeqIO import parse
from numpy.random import random
from tensorflow import Tensor, int64, stack
from tensorflow.sparse import to_dense
from tensorflow.data import TFRecordDataset
from tensorflow.io import (
    FixedLenFeature,
    TFRecordWriter,
    VarLenFeature,
    parse_single_example,
)
from tensorflow.train import Example, Feature, Features, Int64List

from constants import *


def generate_example(sequence, label, base_counts) -> bytes:
    """
    Create a binary-string for each sequence containing the sequence and the bases' counts
    """
    schema = {
        "A_counts": Feature(int64_list=Int64List(value=[sum(base_counts["A"])])),
        "C_counts": Feature(int64_list=Int64List(value=[sum(base_counts["C"])])),
        "G_counts": Feature(int64_list=Int64List(value=[sum(base_counts["G"])])),
        "T_counts": Feature(int64_list=Int64List(value=[sum(base_counts["T"])])),
        "sequence": Feature(int64_list=Int64List(value=encode_sequence(sequence))),
        "label": Feature(int64_list=Int64List(value=encode_sequence(label))),
    }
    example = Example(features=Features(feature=schema))
    return example.SerializeToString()


def encode_sequence(sequence) -> List[int]:
    """
    Encode the DNA sequence using the indices of the BASES constant
    """
    encoded_sequence = [BASES.index(element) for element in sequence]
    return encoded_sequence


def read_fastq(data_file, label_file) -> List[bytes]:
    """
    Parses a data and a label FASTQ files and generates a List of serialized Examples
    """
    examples = []
    with open(data_file) as data, open(label_file) as labels:
        for element, label in zip(parse(data, "fastq"), parse(labels, "fastq")):
            motifs = create([element.seq])
            example = generate_example(
                sequence=str(element.seq),
                label=str(label.seq),
                base_counts=motifs.counts,
            )
            examples.append(example)
    return examples


def create_dataset(
    data_file, label_file, train_eval_test_split=[0.8, 0.1, 0.1]
) -> None:
    """
    Create a training, evaluation and test dataset with a 80/10/30 split respectively
    """
    data = read_fastq(data_file, label_file)
    with TFRecordWriter(TRAIN_DATASET) as training, TFRecordWriter(
        TEST_DATASET
    ) as test, TFRecordWriter(EVAL_DATASET) as evaluation:
        for element in data:
            if random() < train_eval_test_split[0]:
                training.write(element)
            elif random() < train_eval_test_split[0] + train_eval_test_split[1]:
                evaluation.write(element)
            else:
                test.write(element)


def process_input(byte_string) -> Tuple[Tensor, Tensor]:
    """
    Parse a byte-string into an Example object
    """
    schema = {
        "A_counts": FixedLenFeature(shape=[1], dtype=int64),
        "C_counts": FixedLenFeature(shape=[1], dtype=int64),
        "G_counts": FixedLenFeature(shape=[1], dtype=int64),
        "T_counts": FixedLenFeature(shape=[1], dtype=int64),
        "sequence": VarLenFeature(dtype=int64),
        "label": VarLenFeature(dtype=int64),
    }
    parsed_features = parse_single_example(byte_string, features=schema)
    parsed_features["sequence"] = to_dense(parsed_features["sequence"])
    parsed_features["label"] = to_dense(parsed_features["label"])
    features = list(parsed_features.values())[:-1]
    return stack(features, axis=-1), parsed_features["label"]


def read_dataset(filepath) -> TFRecordDataset:
    """
    Read TFRecords files and generate a dataset
    """
    data_input = TFRecordDataset(filenames=filepath)
    dataset = data_input.map(map_func=process_input)
    shuffled_dataset = dataset.shuffle(buffer_size=10000, seed=42)
    batched_dataset = shuffled_dataset.batch(batch_size=BATCH_SIZE).repeat(count=EPOCHS)
    return batched_dataset


def dataset_creation(
    data_file, label_file
) -> Tuple[TFRecordDataset, TFRecordDataset, TFRecordDataset]:
    """
    Generate the TFRecord files and split them into training, validation and test data
    """
    create_dataset(data_file, label_file)
    train_data = read_dataset(TRAIN_DATASET)
    eval_data = read_dataset(EVAL_DATASET)
    test_data = read_dataset(TEST_DATASET)
    return train_data, eval_data, test_data
